#include <stdio.h>
#include <math.h>
#define BUFFSIZE 258

#ifdef UNIX
#ifdef BSD
#include <termios.h>
#elif defined NEXT
#include <malloc.h>
#include <sgtty.h>
#else /* SysV */
#include <malloc.h>
#include <termio.h>
#endif

#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#endif

#if defined(UNIX) && defined(HOTKEYS)

#ifdef BSD
struct termios raw;    /* characteristics for hotkeys */
struct termios noraw;  /* original characteristics */
#elif defined NEXT
struct sgttyb raw;     /* characteristics for hotkeys */
struct sgttyb noraw;   /* original characteristics */
#else /* the System V solution */
struct termio raw;    /* characteristics for hotkeys */
struct termio noraw;  /* original characteristics */
#endif

char onechar[1];      /* the buffer for the hotkey character */
#endif  /* defined(UNIX) ... */

#ifdef UNIX
#define INT32 int
#define SIGINT 2
#elif defined DOS16
#include <stdlib.h>
#include <conio.h>
#define INT32 long
#define REAL float
#define MAXINT 32767
#else
#include <stdlib.h>
#include <conio.h>
#define INT32 int
#define REAL float
#define MAXINT 281474976710655
#endif

typedef struct link
      {
        int nodenum;
        float weight;
        struct link *next;
      } LINK;

typedef struct node
      {
        char name[8];    /* name for the node */
        float sum;
        float value;
        float threshold;
        struct LINK *asc;
      } NODE;

#define linksize sizeof(LINK)
#define nodesize sizeof(NODE)

NODE *units;              /* address of the vector of units */
int nnodes;               /* number of nodes in the network */
int base;                 /* units numbered less than or equal to the */
                          /* base are untouched (frozen) when doing */
                          /* updates */
char outstr[BUFFSIZE];    /* output buffer */
int copyflag;             /* flags making a copy */
FILE *copyfile;           /* file pointer for copy file */
int pagesize;             /* pagesize for paging */
int lineno;               /* counts lines for paging */
char buff[8];             /* buffer for the name of a node */
FILE *data;
char ch;
long seed;

#if defined(UNIX) && defined(HOTKEYS)

#ifdef BSD
void initraw()
{
 int i;
 tcgetattr(0,&noraw);
 raw.c_iflag = noraw.c_iflag;
 raw.c_oflag = noraw.c_oflag;
 raw.c_cflag = noraw.c_cflag;
 raw.c_lflag = noraw.c_lflag & ~ ECHO;
 raw.c_lflag = raw.c_lflag & ~ ICANON;
 for (i=0;i<NCCS;i++) raw.c_cc[i] = noraw.c_cc[i];
 raw.c_cc[VMIN] = 1;
 raw.c_cc[VTIME] = 0;
 raw.c_ispeed = noraw.c_ispeed;
 raw.c_ospeed = noraw.c_ospeed;
}

void setraw()
{
 tcsetattr(0,TCSANOW,&raw);
}

void setnoraw()
{
 tcsetattr(0,TCSANOW,&noraw);
}

#elif defined NEXT
void initraw()
{
 ioctl(0,TIOCGETD,&noraw);
 raw.sg_ispeed = noraw.sg_ispeed;
 raw.sg_ospeed = noraw.sg_ospeed;
 raw.sg_erase = noraw.sg_erase;
 raw.sg_kill = noraw.sg_kill;
 raw.sg_flags = noraw.sg_flags & 0000040;   /* turn on raw mode */
}

void setraw()
{
 ioctl(0,TIOCSETD,&raw);
}

void setnoraw()
{
 ioctl(0,TIOCSETD,&noraw);
}

#else /* System V code */

void initraw()
{
 int i;
 ioctl(0,TCGETA,&noraw);
 raw.c_iflag = noraw.c_iflag;
 raw.c_oflag = noraw.c_oflag;
 raw.c_cflag = noraw.c_cflag;
 raw.c_lflag = noraw.c_lflag & ~ 0x02;   /* hot key bit */
 raw.c_lflag = raw.c_lflag & ~ 010;      /* echo bit */
 raw.c_line = noraw.c_line;
 for (i=0;i<=7;i++) raw.c_cc[i] = noraw.c_cc[i];
 raw.c_cc[4] = 1;
}

void setraw()
{
 ioctl(0,TCSETA,&raw);
}

void setnoraw()
{
 ioctl(0,TCSETA,&noraw);
}
#endif /* end of the System V code */

int getch()
{
 int ch;
 setraw();
 do ch = read(0,onechar,1); while (ch != 1);
 setnoraw();
 return(onechar[0]);
}
#endif  /* end of the if defined(UNIX) ... code */

int pg(char *str) /* paging and making a copy function */
{
char *ch3,action,cr;
int i;
int nbackspaces = 5;

action = 0;
ch3 = str;
while (*ch3 != '\0')
 {
  if (*ch3 == '\n')
   {
    putchar('\n');
    if (copyflag == 1) putc('\n',copyfile);
    lineno = lineno + 1;
    if (pagesize > 0 && lineno > 0 && lineno % pagesize == 0)
     {
      printf("More?");
#ifdef HOTKEYS
      action = getch();
#else
      action = getchar();
      if (action != '\n') cr = getchar();
#endif
      for(i=1;i<=nbackspaces;i++) putchar(8);
      for(i=1;i<=nbackspaces;i++) putchar(' ');
      for(i=1;i<=nbackspaces;i++) putchar(8);
      if (action == 'q') return(1);
      else if (action == 'c') lineno = -MAXINT;
#ifdef UNIX
      else if (action == '\n') lineno = lineno - 1;
#else
      else if (action == 13) lineno = lineno - 1;
#endif
      else if (action == 4) lineno = pagesize / 2;
      else lineno = 0;
     }
   }
  else {putchar(*ch3); if (copyflag == 1) putc(*ch3,copyfile); };
  ch3++;
 };
return(0);
}

#ifndef UNIX
float drand48() {return((rand() & 32767) / 32768.0);}
void srand48(unsigned seed) {srand(seed);}
#endif

void readstr()
{
 int ch, i, j;
 buff[7] = '\0';
 do ch = fgetc(data); while (ch == ' ');
 i = -1;
 do {
    i = i + 1;
    buff[i] = ch;
    ch = fgetc(data);
 } while (i < 6 && (ch != ' ' && ch != '\n'));
 if (i == 6 && (ch != ' ' || ch != '\n')) /* a too long name */
    {
     do ch = fgetc(data); while (ch != ' ' && ch != '\n');
    };
 j = 6;
 if (i < 6)
    do {
       if (i >= 0) buff[j] = buff[i]; else buff[j] = ' ';
       j = j - 1;
       i = i - 1;
    } while (j >= 0);
}

int nodenumber() /* search for the name among the list of nodes */
{
 int i, j, found, notfound;
 i = 1;
 for (i=1;i<=nnodes;i++)
    {
     for (j=0;j<=6;j++) if (buff[j] != units[i].name[j]) break;
     if (j == 7) return(i); /* string found */
    };
 return(0);
}

int scanfordigit()
{
 ch = getc(data);
 while (ch == ' ' || ch == '\n') ch = getc(data);
 if (ch == EOF)
    { 
      pg("unexpected EOF\n");
      exit(2);
    };

 if(ch > '9' || ch < '0')
    {
      pg("digit expected!!!\n");
      if (data != stdin) exit(3);
      return(-1);
    };
 ungetc(ch,data);
 return(1);
};

int rdi()
{
 int temp;
 if (scanfordigit() == -1) return(-1);
 fscanf(data,"%d",&temp);
 return(temp);
}

int randomunit()
{ int u;
  u = (int) (nnodes - base) * drand48() + base + 1;
  return(u);
};

void runrandom(int iter)
{
  int i, from, u;
  float sum, eps;
  LINK *p;

  for(i=1;i<=iter;i++)
     {
       u = randomunit();
       p = (LINK *) units[u].asc;
       sum = 0;
       while (p != NULL)
         {
           from = p->nodenum;
           sum = sum + p->weight * units[from].value;
           p = (LINK *) p->next;
         };
       if (sum >= units[u].threshold)
          eps = sum * (1.0 - units[u].value);
       else
          eps = sum * units[u].value;
       units[u].value = units[u].value + eps;
     };
}

void runnonrandom(int iter)
{
  int i, from, u;
  float sum, eps;
  LINK *p;

  for(i=1;i<=iter;i++)
     {
       for (u=1;u<=nnodes;u++) /* sum activation */
          {
           p = (LINK *) units[u].asc;
           units[u].sum = 0;
           while (p != NULL)
             {
               from = p->nodenum;
               units[u].sum = units[u].sum + p->weight * units[from].value;
               p = (LINK *) p->next;
             };
          };
       for (u=1;u<=nnodes;u++) /* update units */
          {
           if (units[u].sum >= units[u].threshold)
              eps = units[u].sum * (1.0 - units[u].value);
           else
              eps = units[u].sum * units[u].value;
           units[u].value = units[u].value + eps;
          };
     };
}

void status()
{ int i, quarter;

/* prints one column of unit values

   for (i=1;i<=nnodes;i++)
      {
       sprintf(outstr,"%7s %9.5f\n",units[i].name,units[i].value);
       pg(outstr);
      };
*/

/* prints four columns of unit values */

   if (nnodes % 4 == 0) quarter = nnodes / 4;
   else quarter = (nnodes + 3) / 4;
   for (i=1;i<=quarter;i++)
      {
       sprintf(outstr,"%7s %9.5f",units[i].name,units[i].value); pg(outstr);
       if (i + quarter <= nnodes)
          {
           sprintf(outstr,"   %7s %9.5f",units[i+quarter].name,units[i+quarter].value);
           pg(outstr);
          };
       if (i + 2*quarter <= nnodes)
          {
           sprintf(outstr,"   %7s %9.5f",units[i+2* quarter].name,units[i+2*quarter].value);
           pg(outstr);
          };
       if (i + 3*quarter <= nnodes)
          {
           sprintf(outstr,"   %7s %9.5f",units[i+3*quarter].name,units[i+3*quarter].value);
           pg(outstr);
          };
       pg("\n");
      };
}

void cmdloop()
{ int to, i, finished, count, iter, j, nn;
  LINK *p;
  char ch, string[80];
  finished = 0;

 do {
 restart:
      lineno = 0;
      if(data == stdin) pg("[!*cfhlmnqRrSst]? ");
      ch = getc(data);
      while(ch == ' ' || ch == '\n') ch = getc(data);
          switch (ch) {
      case EOF : data = stdin;
                 pg("taking commands from stdin now\n");
                 goto restart;
      case '*' : break;  /* * on a line is a comment */
      case 'c' : fscanf(data,"%d",&copyflag);
                 if (copyflag == 1) copyfile = fopen("copy","w");
                 else if (copyflag == 0) fclose(copyfile);
                 else pg("bad value, use 0 or 1\n");
                 break;
      case 'h' : pg("* on a line is a comment\n");
                 pg("! shell escape\n");
                 pg("c 1 turns on copy function, c 0 turns it off\n");
                 pg("f <int> freezes all units numbered less than or equal to <int>\n");
                 pg("h gives you this help listing\n");
                 pg("l <str1> <str2> <real> creates a link from unit <str1> to unit <str2>\n  with a strength of <real>\n");
                 pg("m <int> makes a network with <int> units\n");
                 pg("n <int> <str> gives unit <int> the name <str>\n");
                 pg("q quits the program\n");
                 pg("R <int> does <int> updates of units chosen AT RANDOM\n");
                 pg("r <int> does <int> passes through all the units in the network\n");
                 pg("S <int> sets the seed\n");
                 pg("s <str> <real> sets unit <str> to <real>\n");
                 pg("t <str> <real> sets a threshold value of <real> for unit <str>\n");
                 break;
/* shell escape */
      case '!' : count = 0;
                 ch = getc(data);
                 while (ch != '\n' && count <= 80)
                    {
                      string[count] = ch;
                      ch = getc(data);
                      count = count + 1;
                    };
                 ungetc(ch,data);
                 string[count] = '\0';
                 system(string);
                 break;
/* run some non-random iterations */
      case 'R' : iter = rdi(); 
                 runrandom(iter);
                 status();
                 ch = '\n';
                 break;
/* run some non-random iterations */
      case 'r' : iter = rdi(); 
                 runnonrandom(iter);
                 status();
                 ch = '\n';
                 break;
/* special command to set up the tic-tac-toe game board, ignore */
      case 'g' : for (i=1;i<=18;i++) fscanf(data,"%f",&units[i].value);
                 for (i=19;i<=50;i++) units[i].value = 0;
                 for (i=51;i<=58;i++) units[i].value = 1;
                 for (i=59;i<=67;i++) units[i].value = 0;
                 break;
/* quit */
      case 'q' : finished = 1;
                 break;
      case 'n':  readstr();
                 nn = nodenumber();
                 if (nn == 0)
                    {sprintf(outstr,"bad label: %s\n",buff); break;};
                 readstr();
                 for (i=0;i<=7;i++) units[nn].name[i] = buff[i];
                 ch = '\n';
                 break;
/* add a link */
      case 'l':  p = (LINK *) malloc(linksize);
                 readstr();
                 p->nodenum = nodenumber();
                 if (p->nodenum == 0)
                    {sprintf(outstr,"bad label: %s\n",buff); break;};
                 readstr();
                 to = nodenumber();
                 if (to == 0)
                    {sprintf(outstr,"bad label: %s\n",buff); break;};
                 fscanf(data,"%f",&p->weight);
                 p->next = units[to].asc;
                 units[to].asc = p;
                 break;
/* make a network (nodes only) */
      case 'm':  fscanf(data,"%d",&nnodes);
                 units = (NODE *) calloc(nnodes+1,nodesize);
                 for (i=1;i<=nnodes;i++)
                    {
                      units[i].sum = 0.0;
                      units[i].asc = NULL;
                      units[i].threshold = 0.0;
                      units[i].value = 0.0;
                      sprintf(units[i].name,"%7d",i);
                    };
                 break;
/* set the threshold of a unit */
      case 't':  readstr();
                 i = nodenumber();
                 if (i == 0)
                    {sprintf(outstr,"bad label: %s\n",buff); break;};
                 fscanf(data,"%f",&units[i].threshold);
                 break;
      case 'f':  fscanf(data,"%d",&base);
                 break;
/* set the value of a unit */
      case 's':  readstr();
                 i = nodenumber();
                 if (i == 0)
                    {sprintf(outstr,"bad label: %s\n",buff); break;};
                 fscanf(data,"%f",&units[i].value);
                 break;
/* set the seed */
      case 'S':  fscanf(data,"%d",&seed);
                 srand48(seed);
                 break;
      default  : pg("unrecognized command; skipping to next line\n");
                 break;
                };
      if (ch != '\n') do ch = getc(data); while (ch != '\n');
   }while (finished == 0);

/* for debugging
  for (i=1;i<=nnodes;i++)
      {
        p = units[i].asc;
        while (p != NULL)
           {
             sprintf(outstr,"%d %d\n",i,p->nodenum);
             pg(outstr);
             p = p->next;
           };
      };
*/
}


void main(int argc,char *argv[])
{
NODE *ptr;
int i;

 setbuf(stdout,NULL);

 /* check for file argument, if any */

 if (argc == 1)
    {
      pg("missing data file name, stdin assumed\n");
      data = stdin;
    }
 else
    if((data = fopen(argv[1],"r")) == (FILE *) NULL)
       {
         sprintf(outstr,"cannot open: %s\n",argv[1]);
         pg(outstr);
         exit(1);
       }

  pagesize = 24;
  srand(0);
#if defined(UNIX) && defined(HOTKEYS)
  initraw();
#endif
  cmdloop();
  if (copyfile != NULL) fclose(copyfile);
}
